@@ -23,12 +23,13 @@
 
 #include "ephy-settings.h"
 
+#include <glib/gi18n.h>
+#include <inttypes.h>
+#include <json-glib/json-glib.h>
 #include <libsoup/soup.h>
 #include <stdio.h>
 #include <string.h>
 
-#define SYNC_ID_LEN 12
-
 static const char hex_digits[] = "0123456789abcdef";
 
 const SecretSchema *
@@ -37,7 +38,7 @@ ephy_sync_utils_get_secret_schema (void)
   static const SecretSchema schema = {
     "org.epiphany.SyncSecrets", SECRET_SCHEMA_NONE,
     {
-      { ACCOUNT_KEY, SECRET_SCHEMA_ATTRIBUTE_STRING },
+      { EPHY_SYNC_SECRET_ACCOUNT_KEY, SECRET_SCHEMA_ATTRIBUTE_STRING },
       { "NULL", 0 },
     }
   };
@@ -219,12 +220,12 @@ ephy_sync_utils_get_random_sync_id (void)
 
   /* The sync id is a base64-urlsafe string. Base64 uses 4 chars to represent 3 bytes,
    * therefore we need ceil(len * 3 / 4) bytes to cover the requested length. */
-  bytes_len = (SYNC_ID_LEN + 3) / 4 * 3;
+  bytes_len = (EPHY_SYNC_BSO_ID_LEN + 3) / 4 * 3;
   bytes = g_malloc (bytes_len);
 
   ephy_sync_utils_generate_random_bytes (NULL, bytes_len, bytes);
   base64 = ephy_sync_utils_base64_urlsafe_encode (bytes, bytes_len, FALSE);
-  id = g_strndup (base64, SYNC_ID_LEN);
+  id = g_strndup (base64, EPHY_SYNC_BSO_ID_LEN);
 
   g_free (base64);
   g_free (bytes);
@@ -232,6 +233,46 @@ ephy_sync_utils_get_random_sync_id (void)
   return id;
 }
 
+char *
+ephy_sync_utils_make_client_record (const char *device_bso_id,
+                                    const char *device_id,
+                                    const char *device_name)
+{
+  JsonNode *node;
+  JsonObject *object;
+  JsonArray *array;
+  char *protocol;
+  char *retval;
+
+  g_assert (device_bso_id);
+  g_assert (device_id);
+  g_assert (device_name);
+
+  array = json_array_new ();
+  protocol = g_strdup_printf ("1.%"PRIu32, EPHY_SYNC_STORAGE_VERSION);
+  json_array_add_string_element (array, protocol);
+
+  object = json_object_new ();
+  json_object_set_string_member (object, "id", device_bso_id);
+  json_object_set_string_member (object, "fxaDeviceId", device_id);
+  json_object_set_string_member (object, "name", device_name);
+  json_object_set_string_member (object, "type", "desktop");
+  json_object_set_string_member (object, "version", VERSION);
+  json_object_set_array_member (object, "protocols", array);
+  json_object_set_string_member (object, "os", "Linux");
+  json_object_set_string_member (object, "appPackage", "org.gnome.epiphany");
+  json_object_set_string_member (object, "application", "Epiphany");
+
+  node = json_node_new (JSON_NODE_OBJECT);
+  json_node_take_object (node, object);
+  retval = json_to_string (node, FALSE);
+
+  g_free (protocol);
+  json_node_unref (node);
+
+  return retval;
+}
+
 void
 ephy_sync_utils_set_device_id (const char *id)
 {
@@ -242,15 +283,25 @@ ephy_sync_utils_set_device_id (const char *id)
 char *
 ephy_sync_utils_get_device_id (void)
 {
-  char *id;
+  return g_settings_get_string (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_DEVICE_ID);
+}
 
-  id = g_settings_get_string (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_DEVICE_ID);
-  if (!g_strcmp0 (id, "")) {
-    g_free (id);
-    id = ephy_sync_utils_get_random_sync_id ();
+char *
+ephy_sync_utils_get_device_bso_id (void)
+{
+  char *device_bso_id;
+  char *device_id;
+
+  device_id = g_settings_get_string (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_DEVICE_ID);
+  if (!g_strcmp0 (device_id, "")) {
+    /* This should never be reached. */
+    return g_strnfill (EPHY_SYNC_BSO_ID_LEN, '0');
   }
 
-  return id;
+  device_bso_id = g_strndup (device_id, EPHY_SYNC_BSO_ID_LEN);
+  g_free (device_id);
+
+  return device_bso_id;
 }
 
 void
@@ -266,11 +317,14 @@ ephy_sync_utils_get_device_name (void)
   char *name;
 
   name = g_settings_get_string (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_DEVICE_NAME);
-  if (!g_strcmp0 (name, "")) {
-    g_free (name);
-    name = g_strdup_printf ("%s's GNOME Web on %s",
-                            g_get_user_name (), g_get_host_name ());
-  }
+  if (g_strcmp0 (name, ""))
+    return name;
+
+  g_free (name);
+  /* Translators: First %s is the name of the user currently logged in on the
+   * machine. The second %s is the machine's name. */
+  name = g_strdup_printf (_("%s’s GNOME Web on %s"), g_get_user_name (), g_get_host_name ());
+  g_settings_set_string (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_DEVICE_NAME, name);
 
   return name;
 }
@@ -341,15 +395,15 @@ ephy_sync_utils_bookmarks_sync_is_enabled (void)
 }
 
 void
-ephy_sync_utils_set_bookmarks_sync_time (double time)
+ephy_sync_utils_set_bookmarks_sync_time (gint64 time)
 {
-  g_settings_set_double (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_BOOKMARKS_TIME, time);
+  g_settings_set_int64 (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_BOOKMARKS_TIME, time);
 }
 
-double
+gint64
 ephy_sync_utils_get_bookmarks_sync_time (void)
 {
-  return g_settings_get_double (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_BOOKMARKS_TIME);
+  return g_settings_get_int64 (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_BOOKMARKS_TIME);
 }
 
 void
@@ -371,15 +425,15 @@ ephy_sync_utils_passwords_sync_is_enabled (void)
 }
 
 void
-ephy_sync_utils_set_passwords_sync_time (double time)
+ephy_sync_utils_set_passwords_sync_time (gint64 time)
 {
-  g_settings_set_double (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_PASSWORDS_TIME, time);
+  g_settings_set_int64 (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_PASSWORDS_TIME, time);
 }
 
-double
+gint64
 ephy_sync_utils_get_passwords_sync_time (void)
 {
-  return g_settings_get_double (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_PASSWORDS_TIME);
+  return g_settings_get_int64 (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_PASSWORDS_TIME);
 }
 
 void
@@ -401,15 +455,15 @@ ephy_sync_utils_history_sync_is_enabled (void)
 }
 
 void
-ephy_sync_utils_set_history_sync_time (double time)
+ephy_sync_utils_set_history_sync_time (gint64 time)
 {
-  g_settings_set_double (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_HISTORY_TIME, time);
+  g_settings_set_int64 (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_HISTORY_TIME, time);
 }
 
-double
+gint64
 ephy_sync_utils_get_history_sync_time (void)
 {
-  return g_settings_get_double (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_HISTORY_TIME);
+  return g_settings_get_int64 (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_HISTORY_TIME);
 }
 
 void
@@ -431,13 +485,13 @@ ephy_sync_utils_open_tabs_sync_is_enabled (void)
 }
 
 void
-ephy_sync_utils_set_open_tabs_sync_time (double time)
+ephy_sync_utils_set_open_tabs_sync_time (gint64 time)
 {
-  g_settings_set_double (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_OPEN_TABS_TIME, time);
+  g_settings_set_int64 (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_OPEN_TABS_TIME, time);
 }
 
-double
+gint64
 ephy_sync_utils_get_open_tabs_sync_time (void)
 {
-  return g_settings_get_double (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_OPEN_TABS_TIME);
+  return g_settings_get_int64 (EPHY_SETTINGS_SYNC, EPHY_PREFS_SYNC_OPEN_TABS_TIME);
 }
