@@ -30,11 +30,11 @@
 
 #include <glib/gi18n.h>
 #include <json-glib/json-glib.h>
+#include <inttypes.h>
 #include <libsoup/soup.h>
+#include <math.h>
 #include <string.h>
 
-#define EPHY_STORAGE_VERSION 5
-
 struct _EphySyncService {
   GObject      parent_instance;
 
@@ -99,8 +99,8 @@ typedef struct {
   char                *endpoint;
   char                *method;
   char                *request_body;
-  double               modified_since;
-  double               unmodified_since;
+  gint64               modified_since;
+  gint64               unmodified_since;
   SoupSessionCallback  callback;
   gpointer             user_data;
 } StorageRequestAsyncData;
@@ -132,12 +132,23 @@ typedef struct {
   EphySynchronizable        *synchronizable;
 } SyncAsyncData;
 
+typedef struct {
+  EphySyncService           *service;
+  EphySynchronizableManager *manager;
+  GPtrArray                 *synchronizables;
+  guint                      start;
+  guint                      end;
+  char                      *batch_id;
+  gboolean                   batch_is_last;
+  gboolean                   sync_done;
+} BatchUploadAsyncData;
+
 static StorageRequestAsyncData *
 storage_request_async_data_new (const char          *endpoint,
                                 const char          *method,
                                 const char          *request_body,
-                                double               modified_since,
-                                double               unmodified_since,
+                                gint64               modified_since,
+                                gint64               unmodified_since,
                                 SoupSessionCallback  callback,
                                 gpointer             user_data)
 {
@@ -270,6 +281,54 @@ sync_async_data_free (SyncAsyncData *data)
   g_slice_free (SyncAsyncData, data);
 }
 
+static inline BatchUploadAsyncData *
+batch_upload_async_data_new (EphySyncService           *service,
+                             EphySynchronizableManager *manager,
+                             GPtrArray                 *synchronizables,
+                             guint                      start,
+                             guint                      end,
+                             const char                *batch_id,
+                             gboolean                   batch_is_last,
+                             gboolean                   sync_done)
+{
+  BatchUploadAsyncData *data;
+
+  data = g_slice_new (BatchUploadAsyncData);
+  data->service = g_object_ref (service);
+  data->manager = g_object_ref (manager);
+  data->synchronizables = g_ptr_array_ref (synchronizables);
+  data->start = start;
+  data->end = end;
+  data->batch_id = g_strdup (batch_id);
+  data->batch_is_last = batch_is_last;
+  data->sync_done = sync_done;
+
+  return data;
+}
+
+static inline BatchUploadAsyncData *
+batch_upload_async_data_dup (BatchUploadAsyncData *data)
+{
+  g_assert (data);
+
+  return batch_upload_async_data_new (data->service, data->manager,
+                                      data->synchronizables, data->start,
+                                      data->end, data->batch_id,
+                                      data->batch_is_last, data->sync_done);
+}
+
+static inline void
+batch_upload_async_data_free (BatchUploadAsyncData *data)
+{
+  g_assert (data);
+
+  g_object_unref (data->service);
+  g_object_unref (data->manager);
+  g_ptr_array_unref (data->synchronizables);
+  g_free (data->batch_id);
+  g_slice_free (BatchUploadAsyncData, data);
+}
+
 static void
 ephy_sync_service_set_property (GObject      *object,
                                 guint         prop_id,
@@ -406,7 +465,7 @@ ephy_sync_service_fxa_hawk_post (EphySyncService     *self,
   g_assert (key);
   g_assert (request_body);
 
-  url = g_strdup_printf ("%s/%s", FIREFOX_ACCOUNTS_SERVER_URL, endpoint);
+  url = g_strdup_printf ("%s/%s", EPHY_SYNC_FX_ACCOUNTS_SERVER_URL, endpoint);
   msg = soup_message_new (SOUP_METHOD_POST, url);
   soup_message_set_request (msg, content_type, SOUP_MEMORY_COPY,
                             request_body, strlen (request_body));
@@ -442,7 +501,7 @@ ephy_sync_service_fxa_hawk_get (EphySyncService     *self,
   g_assert (id);
   g_assert (key);
 
-  url = g_strdup_printf ("%s/%s", FIREFOX_ACCOUNTS_SERVER_URL, endpoint);
+  url = g_strdup_printf ("%s/%s", EPHY_SYNC_FX_ACCOUNTS_SERVER_URL, endpoint);
   msg = soup_message_new (SOUP_METHOD_GET, url);
   header = ephy_sync_crypto_hawk_header_new (url, "GET", id, key, key_len, NULL);
   soup_message_headers_append (msg->request_headers, "authorization", header->header);
@@ -478,16 +537,16 @@ ephy_sync_service_send_storage_request (EphySyncService         *self,
                               data->request_body, strlen (data->request_body));
   }
 
-  if (!g_strcmp0 (data->method, SOUP_METHOD_PUT))
+  if (!g_strcmp0 (data->method, SOUP_METHOD_PUT) || !g_strcmp0 (data->method, SOUP_METHOD_POST))
     soup_message_headers_append (msg->request_headers, "content-type", content_type);
 
   if (data->modified_since >= 0) {
-    if_modified_since = g_strdup_printf ("%.2lf", data->modified_since);
+    if_modified_since = g_strdup_printf ("%"PRId64, data->modified_since);
     soup_message_headers_append (msg->request_headers, "X-If-Modified-Since", if_modified_since);
   }
 
   if (data->unmodified_since >= 0) {
-    if_unmodified_since = g_strdup_printf ("%.2lf", data->unmodified_since);
+    if_unmodified_since = g_strdup_printf ("%"PRId64, data->unmodified_since);
     soup_message_headers_append (msg->request_headers, "X-If-Unmodified-Since", if_unmodified_since);
   }
 
@@ -596,7 +655,7 @@ ephy_sync_service_verify_certificate (EphySyncService *self,
     g_warning ("JSON object has missing or invalid 'email' member");
     goto out;
   }
-  uri = soup_uri_new (FIREFOX_ACCOUNTS_SERVER_URL);
+  uri = soup_uri_new (EPHY_SYNC_FX_ACCOUNTS_SERVER_URL);
   expected = g_strdup_printf ("%s@%s",
                               ephy_sync_service_get_secret (self, secrets[UID]),
                               soup_uri_get_host (uri));
@@ -645,7 +704,7 @@ ephy_sync_service_forget_secrets (EphySyncService *self)
   g_return_if_fail (user);
 
   attributes = secret_attributes_build (EPHY_SYNC_SECRET_SCHEMA,
-                                        ACCOUNT_KEY, user,
+                                        EPHY_SYNC_SECRET_ACCOUNT_KEY, user,
                                         NULL);
   secret_service_clear (NULL, EPHY_SYNC_SECRET_SCHEMA, attributes, NULL,
                         (GAsyncReadyCallback)forget_secrets_cb, NULL);
@@ -688,7 +747,8 @@ ephy_sync_service_destroy_session (EphySyncService *self,
     session_token = ephy_sync_service_get_secret (self, secrets[SESSION_TOKEN]);
   g_assert (session_token);
 
-  url = g_strdup_printf ("%s/session/destroy", FIREFOX_ACCOUNTS_SERVER_URL);
+  /* This also destroys the device associated with the session token. */
+  url = g_strdup_printf ("%s/session/destroy", EPHY_SYNC_FX_ACCOUNTS_SERVER_URL);
   ephy_sync_crypto_derive_session_token (session_token, &token_id,
                                          &req_hmac_key, &tmp);
   token_id_hex = ephy_sync_utils_encode_hex (token_id, 32);
@@ -817,7 +877,7 @@ ephy_sync_service_trade_browserid_assertion (EphySyncService *self)
   g_assert (self->certificate);
   g_assert (self->key_pair);
 
-  audience = ephy_sync_utils_get_audience (TOKEN_SERVER_URL);
+  audience = ephy_sync_utils_get_audience (EPHY_SYNC_FX_TOKEN_SERVER_URL);
   assertion = ephy_sync_crypto_create_assertion (self->certificate, audience,
                                                  300, self->key_pair);
   kb = ephy_sync_utils_decode_hex (ephy_sync_service_get_secret (self, secrets[MASTER_KEY]));
@@ -825,7 +885,7 @@ ephy_sync_service_trade_browserid_assertion (EphySyncService *self)
   client_state = g_strndup (hashed_kb, 32);
   authorization = g_strdup_printf ("BrowserID %s", assertion);
 
-  msg = soup_message_new (SOUP_METHOD_GET, TOKEN_SERVER_URL);
+  msg = soup_message_new (SOUP_METHOD_GET, EPHY_SYNC_FX_TOKEN_SERVER_URL);
   /* We need to add the X-Client-State header so that the Token Server will
    * recognize accounts that were previously used to sync Firefox data too.
    */
@@ -891,7 +951,7 @@ get_signed_certificate_cb (SoupSession *session,
   if (json_object_get_int_member (json, "errno") == 110) {
     message = _("The password of your Firefox account seems to have been changed.");
     suggestion = _("Please visit Preferences and sign in with the new password to continue syncing.");
-    ephy_sync_service_sign_out (self, FALSE);
+    ephy_sync_service_sign_out (self);
   }
 
   g_warning ("Failed to sign certificate. Status code: %u, response: %s",
@@ -982,8 +1042,8 @@ ephy_sync_service_queue_storage_request (EphySyncService     *self,
                                          const char          *endpoint,
                                          const char          *method,
                                          const char          *request_body,
-                                         double               modified_since,
-                                         double               unmodified_since,
+                                         gint64               modified_since,
+                                         gint64               unmodified_since,
                                          SoupSessionCallback  callback,
                                          gpointer             user_data)
 {
@@ -1178,7 +1238,7 @@ upload_synchronizable_cb (SoupSession *session,
                           gpointer     user_data)
 {
   SyncAsyncData *data = (SyncAsyncData *)user_data;
-  double time_modified;
+  gint64 time_modified;
 
   /* Code 412 means that there is a more recent version on the server.
    * Download it.
@@ -1188,7 +1248,7 @@ upload_synchronizable_cb (SoupSession *session,
     ephy_sync_service_download_synchronizable (data->service, data->manager, data->synchronizable);
   } else if (msg->status_code == 200) {
     LOG ("Successfully uploaded to server");
-    time_modified = g_ascii_strtod (msg->response_body->data, NULL);
+    time_modified = ceil (g_ascii_strtod (msg->response_body->data, NULL));
     ephy_synchronizable_set_server_time_modified (data->synchronizable, time_modified);
     ephy_synchronizable_manager_save (data->manager, data->synchronizable);
   } else {
@@ -1213,7 +1273,7 @@ ephy_sync_service_upload_synchronizable (EphySyncService           *self,
   char *id_safe;
   const char *collection;
   const char *id;
-  double time_modified;
+  gint64 time_modified;
 
   g_assert (EPHY_IS_SYNC_SERVICE (self));
   g_assert (EPHY_IS_SYNCHRONIZABLE_MANAGER (manager));
@@ -1245,22 +1305,193 @@ ephy_sync_service_upload_synchronizable (EphySyncService           *self,
   ephy_sync_crypto_key_bundle_free (bundle);
 }
 
-static void
-merge_collection_finished_cb (GList    *to_upload,
-                              gboolean  should_force,
-                              gpointer  user_data)
+static GPtrArray *
+ephy_sync_service_split_into_batches (EphySyncService           *self,
+                                      EphySynchronizableManager *manager,
+                                      GPtrArray                 *synchronizables,
+                                      guint                      start,
+                                      guint                      end)
 {
-  SyncCollectionAsyncData *data = (SyncCollectionAsyncData *)user_data;
+  SyncCryptoKeyBundle *bundle;
+  GPtrArray *batches;
+  const char *collection;
 
-  for (GList *l = to_upload; l && l->data; l = l->next)
-    ephy_sync_service_upload_synchronizable (data->service, data->manager,
-                                             l->data, should_force);
+  g_assert (EPHY_IS_SYNC_SERVICE (self));
+  g_assert (EPHY_IS_SYNCHRONIZABLE_MANAGER (manager));
+  g_assert (synchronizables);
 
-  if (data->is_last)
+  batches = g_ptr_array_new_with_free_func (g_free);
+  collection = ephy_synchronizable_manager_get_collection_name (manager);
+  bundle = ephy_sync_service_get_key_bundle (self, collection);
+
+  for (guint i = start; i < end; i += EPHY_SYNC_BATCH_SIZE) {
+    JsonNode *node = json_node_new (JSON_NODE_ARRAY);
+    JsonArray *array = json_array_new ();
+
+    for (guint k = i; k < MIN (i + EPHY_SYNC_BATCH_SIZE, end); k++) {
+      EphySynchronizable *synchronizable = g_ptr_array_index (synchronizables, k);
+      JsonNode *bso = ephy_synchronizable_to_bso (synchronizable, bundle);
+      JsonObject *object = json_object_ref (json_node_get_object (bso));
+
+      json_array_add_object_element (array, object);
+      json_node_unref (bso);
+    }
+
+    json_node_take_array (node, array);
+    g_ptr_array_add (batches, json_to_string (node, FALSE));
+    json_node_unref (node);
+  }
+
+  ephy_sync_crypto_key_bundle_free (bundle);
+
+  return batches;
+}
+
+static void
+commit_batch_cb (SoupSession *session,
+                 SoupMessage *msg,
+                 gpointer     user_data)
+{
+  BatchUploadAsyncData *data = user_data;
+  const char *last_modified;
+
+  if (msg->status_code != 200) {
+    g_warning ("Failed to commit batch. Status code: %u, response: %s",
+               msg->status_code, msg->response_body->data);
+  } else {
+    LOG ("Successfully committed batches");
+    /* Update sync time. */
+    last_modified = soup_message_headers_get_one (msg->response_headers, "X-Last-Modified");
+    ephy_synchronizable_manager_set_sync_time (data->manager, g_ascii_strtod (last_modified, NULL));
+  }
+
+  if (data->sync_done)
     g_signal_emit (data->service, signals[SYNC_FINISHED], 0);
+  batch_upload_async_data_free (data);
+}
 
-  if (to_upload)
-    g_list_free_full (to_upload, g_object_unref);
+static void
+upload_batch_cb (SoupSession *session,
+                 SoupMessage *msg,
+                 gpointer     user_data)
+{
+  BatchUploadAsyncData *data = user_data;
+  const char *collection;
+  char *endpoint = NULL;
+
+  /* Note: "202 Accepted" status code. */
+  if (msg->status_code != 202) {
+    g_warning ("Failed to upload batch. Status code: %u, response: %s",
+               msg->status_code, msg->response_body->data);
+  } else {
+    LOG ("Successfully uploaded batch");
+  }
+
+  if (!data->batch_is_last)
+    goto out;
+
+  collection = ephy_synchronizable_manager_get_collection_name (data->manager);
+  endpoint = g_strdup_printf ("storage/%s?commit=true&batch=%s", collection, data->batch_id);
+  ephy_sync_service_queue_storage_request (data->service, endpoint,
+                                           SOUP_METHOD_POST, "[]", -1, -1,
+                                           commit_batch_cb,
+                                           batch_upload_async_data_dup (data));
+
+out:
+  g_free (endpoint);
+  /* Remove last reference to the array with the items to upload. */
+  if (data->batch_is_last)
+    g_ptr_array_unref (data->synchronizables);
+  batch_upload_async_data_free (data);
+}
+
+static void
+start_batch_upload_cb (SoupSession *session,
+                       SoupMessage *msg,
+                       gpointer     user_data)
+{
+  BatchUploadAsyncData *data = user_data;
+  GPtrArray *batches = NULL;
+  JsonNode *node = NULL;
+  JsonObject *object;
+  GError *error = NULL;
+  const char *collection;
+  char *endpoint = NULL;
+
+  /* Note: "202 Accepted" status code. */
+  if (msg->status_code != 202) {
+    g_warning ("Failed to start batch upload. Status code: %u, response: %s",
+               msg->status_code, msg->response_body->data);
+    goto out;
+  }
+
+  node = json_from_string (msg->response_body->data, &error);
+  if (error) {
+    g_warning ("Response is not a valid JSON: %s", error->message);
+    g_error_free (error);
+    goto out;
+  }
+
+  object = json_node_get_object (node);
+  data->batch_id = soup_uri_encode (json_object_get_string_member (object, "batch"), NULL);
+  collection = ephy_synchronizable_manager_get_collection_name (data->manager);
+  endpoint = g_strdup_printf ("storage/%s?batch=%s", collection, data->batch_id);
+
+  batches = ephy_sync_service_split_into_batches (data->service, data->manager,
+                                                  data->synchronizables,
+                                                  data->start, data->end);
+  for (guint i = 0; i < batches->len; i++) {
+    BatchUploadAsyncData *data_dup = batch_upload_async_data_dup (data);
+
+    if (i == batches->len - 1)
+      data_dup->batch_is_last = TRUE;
+
+    ephy_sync_service_queue_storage_request (data->service, endpoint, SOUP_METHOD_POST,
+                                             g_ptr_array_index (batches, i), -1, -1,
+                                             upload_batch_cb, data_dup);
+  }
+
+out:
+  g_free (endpoint);
+  if (node)
+    json_node_unref (node);
+  if (batches)
+    g_ptr_array_unref (batches);
+  batch_upload_async_data_free (data);
+}
+
+static void
+merge_collection_finished_cb (GPtrArray *to_upload,
+                              gpointer   user_data)
+{
+  SyncCollectionAsyncData *data = user_data;
+  BatchUploadAsyncData *bdata;
+  guint step = EPHY_SYNC_MAX_BATCHES * EPHY_SYNC_BATCH_SIZE;
+  const char *collection;
+  char *endpoint = NULL;
+
+  if (!to_upload || to_upload->len == 0) {
+    if (data->is_last)
+      g_signal_emit (data->service, signals[SYNC_FINISHED], 0);
+    goto out;
+  }
+
+  collection = ephy_synchronizable_manager_get_collection_name (data->manager);
+  endpoint = g_strdup_printf ("storage/%s?batch=true", collection);
+
+  for (guint i = 0; i < to_upload->len; i += step) {
+    bdata = batch_upload_async_data_new (data->service, data->manager,
+                                         to_upload, i,
+                                         MIN (i + step, to_upload->len),
+                                         NULL, FALSE,
+                                         data->is_last && i + step >= to_upload->len);
+    ephy_sync_service_queue_storage_request (data->service, endpoint,
+                                             SOUP_METHOD_POST, "[]", -1, -1,
+                                             start_batch_upload_cb, bdata);
+  }
+
+out:
+  g_free (endpoint);
   sync_collection_async_data_free (data);
 }
 
@@ -1277,7 +1508,6 @@ sync_collection_cb (SoupSession *session,
   GError *error = NULL;
   GType type;
   const char *collection;
-  const char *last_modified;
   gboolean is_deleted;
 
   collection = ephy_synchronizable_manager_get_collection_name (data->manager);
@@ -1318,11 +1548,7 @@ sync_collection_cb (SoupSession *session,
        g_list_length (data->remotes_updated),
        collection);
 
-  /* Update sync time. */
-  last_modified = soup_message_headers_get_one (msg->response_headers, "X-Last-Modified");
-  ephy_synchronizable_manager_set_sync_time (data->manager, g_ascii_strtod (last_modified, NULL));
   ephy_synchronizable_manager_set_is_initial_sync (data->manager, FALSE);
-
   ephy_synchronizable_manager_merge (data->manager, data->is_initial,
                                      data->remotes_deleted, data->remotes_updated,
                                      merge_collection_finished_cb, data);
@@ -1359,7 +1585,7 @@ ephy_sync_service_sync_collection (EphySyncService           *self,
   if (is_initial) {
     endpoint = g_strdup_printf ("storage/%s?full=true", collection);
   } else {
-    endpoint = g_strdup_printf ("storage/%s?newer=%.2lf&full=true", collection,
+    endpoint = g_strdup_printf ("storage/%s?newer=%"PRId64"&full=true", collection,
                                 ephy_synchronizable_manager_get_sync_time (manager));
   }
 
@@ -1518,7 +1744,7 @@ ephy_sync_service_load_secrets (EphySyncService *self)
 
   user = ephy_sync_utils_get_sync_user ();
   attributes = secret_attributes_build (EPHY_SYNC_SECRET_SCHEMA,
-                                        ACCOUNT_KEY, user,
+                                        EPHY_SYNC_SECRET_ACCOUNT_KEY, user,
                                         NULL);
   secret_service_search (NULL, EPHY_SYNC_SECRET_SCHEMA, attributes,
                          SECRET_SEARCH_UNLOCK | SECRET_SEARCH_LOAD_SECRETS,
@@ -1580,7 +1806,7 @@ ephy_sync_service_store_secrets (EphySyncService *self)
 
   secret = secret_value_new (json_string, -1, "text/plain");
   attributes = secret_attributes_build (EPHY_SYNC_SECRET_SCHEMA,
-                                        ACCOUNT_KEY, self->user,
+                                        EPHY_SYNC_SECRET_ACCOUNT_KEY, self->user,
                                         NULL);
   /* Translators: %s is the email of the user. */
   label = g_strdup_printf (_("The sync secrets of %s"), self->user);
@@ -1598,6 +1824,79 @@ ephy_sync_service_store_secrets (EphySyncService *self)
   json_node_unref (node);
 }
 
+static void
+upload_client_record_cb (SoupSession *session,
+                         SoupMessage *msg,
+                         gpointer     user_data)
+{
+  EphySyncService *self = EPHY_SYNC_SERVICE (user_data);
+
+  if (msg->status_code != 200) {
+    g_warning ("Failed to upload client record. Status code: %u, response: %s",
+               msg->status_code, msg->response_body->data);
+    if (self->is_signing_in)
+      ephy_sync_service_report_sign_in_error (self, _("Failed to upload client record."), NULL, TRUE);
+  } else {
+    LOG ("Successfully uploaded client record");
+    if (self->is_signing_in)
+      ephy_sync_service_store_secrets (self);
+  }
+}
+
+static void
+ephy_sync_service_upload_client_record (EphySyncService *self)
+{
+  SyncCryptoKeyBundle *bundle;
+  JsonNode *node;
+  JsonObject *bso;
+  char *device_bso_id;
+  char *device_id;
+  char *device_name;
+  char *record;
+  char *encrypted;
+  char *body;
+  char *endpoint;
+
+  g_assert (EPHY_IS_SYNC_SERVICE (self));
+
+  /* Make device ID and name. */
+  device_bso_id = ephy_sync_utils_get_device_bso_id ();
+  device_id = ephy_sync_utils_get_device_id ();
+  device_name = ephy_sync_utils_get_device_name ();
+
+  /* Make BSO as string. */
+  record = ephy_sync_utils_make_client_record (device_bso_id, device_id, device_name);
+  bundle = ephy_sync_service_get_key_bundle (self, "clients");
+  encrypted = ephy_sync_crypto_encrypt_record (record, bundle);
+
+  bso = json_object_new ();
+  json_object_set_string_member (bso, "id", device_bso_id);
+  json_object_set_string_member (bso, "payload", encrypted);
+
+  node = json_node_new (JSON_NODE_OBJECT);
+  json_node_set_object (node, bso);
+  body = json_to_string (node, FALSE);
+
+  /* Upload BSO and store the new device ID and name. */
+  LOG ("Uploading client record, device_bso_id=%s, device_id=%s, device_name=%s",
+       device_bso_id, device_id, device_name);
+  endpoint = g_strdup_printf ("storage/clients/%s", device_bso_id);
+  ephy_sync_service_queue_storage_request (self, endpoint,
+                                           SOUP_METHOD_PUT, body, -1, -1,
+                                           upload_client_record_cb, self);
+
+  g_free (device_bso_id);
+  g_free (device_id);
+  g_free (device_name);
+  g_free (record);
+  g_free (encrypted);
+  g_free (endpoint);
+  g_free (body);
+  json_object_unref (bso);
+  json_node_unref (node);
+  ephy_sync_crypto_key_bundle_free (bundle);
+}
+
 static void
 ephy_sync_service_finalize (GObject *object)
 {
@@ -1731,7 +2030,7 @@ upload_crypto_keys_cb (SoupSession *session,
   } else {
     LOG ("Successfully uploaded crypto/keys record");
     ephy_sync_service_set_secret (self, secrets[CRYPTO_KEYS], self->crypto_keys);
-    ephy_sync_service_register_device (self, NULL);
+    ephy_sync_service_upload_client_record (self);
   }
 
   g_clear_pointer (&self->crypto_keys, g_free);
@@ -1829,7 +2128,7 @@ get_crypto_keys_cb (SoupSession *session,
   }
 
   ephy_sync_service_set_secret (self, secrets[CRYPTO_KEYS], crypto_keys);
-  ephy_sync_service_register_device (self, NULL);
+  ephy_sync_service_upload_client_record (self);
   goto out_no_error;
 
 out_error:
@@ -1921,7 +2220,7 @@ ephy_sync_service_upload_meta_global (EphySyncService *self)
   json_object_set_object_member (engines, "tabs", make_engine_object (1));
   json_object_set_object_member (engines, "forms", make_engine_object (1));
   json_object_set_object_member (payload, "engines", engines);
-  json_object_set_int_member (payload, "storageVersion", EPHY_STORAGE_VERSION);
+  json_object_set_int_member (payload, "storageVersion", EPHY_SYNC_STORAGE_VERSION);
   sync_id = ephy_sync_utils_get_random_sync_id ();
   json_object_set_string_member (payload, "syncID", sync_id);
   json_node_set_object (node, payload);
@@ -1999,11 +2298,11 @@ verify_storage_version_cb (SoupSession *session,
     goto out_error;
   }
   storage_version = json_object_get_int_member (json, "storageVersion");
-  if (storage_version != EPHY_STORAGE_VERSION) {
+  if (storage_version != EPHY_SYNC_STORAGE_VERSION) {
     /* Translators: the %d is the storage version, the \n is a newline character. */
     message = g_strdup_printf (_("Your Firefox Account uses storage version %d. "
                                  "Web only supports version %d."),
-                               EPHY_STORAGE_VERSION,
+                               EPHY_SYNC_STORAGE_VERSION,
                                storage_version);
     goto out_error;
   }
@@ -2034,6 +2333,94 @@ ephy_sync_service_verify_storage_version (EphySyncService *self)
                                            verify_storage_version_cb, self);
 }
 
+static void
+upload_fxa_device_cb (SoupSession *session,
+                      SoupMessage *msg,
+                      gpointer     user_data)
+{
+  EphySyncService *self = user_data;
+  JsonNode *node;
+  JsonObject *object;
+  GError *error = NULL;
+
+  if (msg->status_code != 200) {
+    g_warning ("Failed to upload device info on FxA Server. Status code: %u, response: %s",
+               msg->status_code, msg->response_body->data);
+    goto out_error;
+  }
+
+  node = json_from_string (msg->response_body->data, &error);
+  if (error) {
+    g_warning ("Response is not a valid JSON: %s", error->message);
+    g_error_free (error);
+    goto out_error;
+  }
+
+  object = json_node_get_object (node);
+  ephy_sync_utils_set_device_id (json_object_get_string_member (object, "id"));
+  json_node_unref (node);
+
+  LOG ("Successfully uploaded device info on FxA Server");
+  if (self->is_signing_in)
+    ephy_sync_service_verify_storage_version (self);
+  return;
+
+out_error:
+  if (self->is_signing_in)
+    ephy_sync_service_report_sign_in_error (self, _("Failed to upload device info"), NULL, TRUE);
+}
+
+static void
+ephy_sync_service_upload_fxa_device (EphySyncService *self)
+{
+  JsonNode *node;
+  JsonObject *object;
+  const char *session_token;
+  char *body;
+  char *device_name;
+  char *token_id_hex;
+  guint8 *token_id;
+  guint8 *req_hmac_key;
+  guint8 *tmp;
+
+  g_assert (EPHY_IS_SYNC_SERVICE (self));
+
+  object = json_object_new ();
+  device_name = ephy_sync_utils_get_device_name ();
+  json_object_set_string_member (object, "name", device_name);
+  json_object_set_string_member (object, "type", "desktop");
+
+  /* If we are signing in, the ID for the newly registered device will be returned
+   * by the FxA server in the response. Otherwise, we are updating the current
+   * device (i.e. setting its name), so we use the previously obtained ID. */
+  if (!self->is_signing_in) {
+    char *device_id = ephy_sync_utils_get_device_id ();
+    json_object_set_string_member (object, "id", device_id);
+    g_free (device_id);
+  }
+
+  node = json_node_new (JSON_NODE_OBJECT);
+  json_node_take_object (node, object);
+  body = json_to_string (node, FALSE);
+
+  session_token = ephy_sync_service_get_secret (self, secrets[SESSION_TOKEN]);
+  ephy_sync_crypto_derive_session_token (session_token, &token_id, &req_hmac_key, &tmp);
+  token_id_hex = ephy_sync_utils_encode_hex (token_id, 32);
+
+  LOG ("Uploading device info on FxA Server...");
+  ephy_sync_service_fxa_hawk_post (self, "account/device", token_id_hex,
+                                   req_hmac_key, 32, body,
+                                   upload_fxa_device_cb, self);
+
+  g_free (body);
+  g_free (device_name);
+  g_free (token_id_hex);
+  g_free (token_id);
+  g_free (req_hmac_key);
+  g_free (tmp);
+  json_node_unref (node);
+}
+
 static void
 ephy_sync_service_sign_in_finish (EphySyncService *self,
                                   SignInAsyncData *data,
@@ -2072,7 +2459,7 @@ ephy_sync_service_sign_in_finish (EphySyncService *self,
   kb_hex = ephy_sync_utils_encode_hex (kb, 32);
   ephy_sync_service_set_secret (self, secrets[MASTER_KEY], kb_hex);
 
-  ephy_sync_service_verify_storage_version (self);
+  ephy_sync_service_upload_fxa_device (self);
 
   g_free (kb_hex);
   g_free (kb);
@@ -2266,105 +2653,16 @@ ephy_sync_service_unregister_manager (EphySyncService           *self,
   g_signal_handlers_disconnect_by_func (manager, synchronizable_modified_cb, self);
 }
 
-static void
-register_device_cb (SoupSession *session,
-                    SoupMessage *msg,
-                    gpointer     user_data)
-{
-  EphySyncService *self = EPHY_SYNC_SERVICE (user_data);
-
-  if (msg->status_code != 200) {
-    g_warning ("Failed to register device. Status code: %u, response: %s",
-               msg->status_code, msg->response_body->data);
-    ephy_sync_service_report_sign_in_error (self,
-                                            _("Failed to register device."),
-                                            NULL, TRUE);
-  } else {
-    LOG ("Successfully registered device");
-    if (self->is_signing_in)
-      ephy_sync_service_store_secrets (self);
-  }
-}
-
 void
-ephy_sync_service_register_device (EphySyncService *self,
-                                   const char      *device_name)
+ephy_sync_service_update_device_name (EphySyncService *self,
+                                      const char      *name)
 {
-  SyncCryptoKeyBundle *bundle;
-  JsonNode *node;
-  JsonObject *bso;
-  JsonObject *record;
-  JsonArray *array;
-  char *id;
-  char *name;
-  char *protocol;
-  char *cleartext;
-  char *encrypted;
-  char *body;
-  char *endpoint;
+  g_assert (EPHY_IS_SYNC_SERVICE (self));
+  g_assert (name);
 
-  g_return_if_fail (EPHY_IS_SYNC_SERVICE (self));
-
-  /* Make protocol. */
-  protocol = g_strdup_printf ("1.%d", EPHY_STORAGE_VERSION);
-  array = json_array_new ();
-  json_array_add_string_element (array, protocol);
-
-  /* Make device ID and name. */
-  id = ephy_sync_utils_get_device_id ();
-  if (device_name)
-    name = g_strdup (device_name);
-  else
-    name = ephy_sync_utils_get_device_name ();
-
-  /* Set record members. */
-  record = json_object_new ();
-  json_object_set_string_member (record, "id", id);
-  json_object_set_string_member (record, "fxaDeviceId",
-                                 ephy_sync_service_get_secret (self, secrets[UID]));
-  json_object_set_string_member (record, "name", name);
-  json_object_set_string_member (record, "type", "desktop");
-  json_object_set_string_member (record, "version", VERSION);
-  json_object_set_array_member (record, "protocols", array);
-  json_object_set_string_member (record, "os", "Linux");
-  json_object_set_string_member (record, "appPackage", "org.gnome.epiphany");
-  json_object_set_string_member (record, "application", "Epiphany");
-
-  /* Get record's string representation. */
-  node = json_node_new (JSON_NODE_OBJECT);
-  json_node_set_object (node, record);
-  cleartext = json_to_string (node, FALSE);
-
-  /* Make BSO as string. */
-  bundle = ephy_sync_service_get_key_bundle (self, "clients");
-  encrypted = ephy_sync_crypto_encrypt_record (cleartext, bundle);
-  bso = json_object_new ();
-  json_object_set_string_member (bso, "id", id);
-  json_object_set_string_member (bso, "payload", encrypted);
-  json_node_set_object (node, bso);
-  body = json_to_string (node, FALSE);
-
-  /* Upload BSO and store the new device ID and name. */
-  LOG ("Registering device with name '%s'...", name);
-  endpoint = g_strdup_printf ("storage/clients/%s", id);
-  ephy_sync_service_queue_storage_request (self, endpoint,
-                                           SOUP_METHOD_PUT, body, -1, -1,
-                                           register_device_cb, self);
-
-  ephy_sync_utils_set_device_id (id);
   ephy_sync_utils_set_device_name (name);
-
-  g_free (endpoint);
-  g_free (body);
-  g_free (encrypted);
-  ephy_sync_crypto_key_bundle_free (bundle);
-  g_free (cleartext);
-  g_free (name);
-  g_free (id);
-  g_free (protocol);
-  json_object_unref (record);
-  json_object_unref (bso);
-  json_node_unref (node);
+  ephy_sync_service_upload_fxa_device (self);
+  ephy_sync_service_upload_client_record (self);
 }
 
 static void
@@ -2372,75 +2670,80 @@ delete_open_tabs_record_cb (SoupSession *session,
                             SoupMessage *msg,
                             gpointer     user_data)
 {
+  EphySyncService *self = EPHY_SYNC_SERVICE (user_data);
+  const char *session_token;
+
   if (msg->status_code != 200) {
     g_warning ("Failed to delete open tabs record. Status code: %u, response: %s",
                msg->status_code, msg->response_body->data);
   } else {
     LOG ("Successfully deleted open tabs record");
   }
+
+  ephy_sync_service_clear_storage_queue (self);
+  ephy_sync_service_clear_storage_credentials (self);
+
+  session_token = ephy_sync_service_get_secret (self, secrets[SESSION_TOKEN]);
+  ephy_sync_service_destroy_session (self, session_token);
+
+  ephy_sync_service_forget_secrets (self);
+  ephy_sync_utils_set_device_id (NULL);
+  ephy_sync_utils_set_sync_user (NULL);
 }
 
 static void
-unregister_device_cb (SoupSession *session,
-                      SoupMessage *msg,
-                      gpointer     user_data)
+delete_client_record_cb (SoupSession *session,
+                         SoupMessage *msg,
+                         gpointer     user_data)
 {
+  EphySyncService *self = EPHY_SYNC_SERVICE (user_data);
+  char *endpoint;
+  char *device_bso_id;
+
   if (msg->status_code != 200) {
-    g_warning ("Failed to unregister device. Status code: %u, response: %s",
+    g_warning ("Failed to delete client record. Status code: %u, response: %s",
                msg->status_code, msg->response_body->data);
   } else {
-    LOG ("Successfully unregistered device");
+    LOG ("Successfully deleted client record");
   }
+
+  device_bso_id = ephy_sync_utils_get_device_bso_id ();
+  /* Delete the open tabs record associated to this device. */
+  endpoint = g_strdup_printf ("storage/tabs/%s", device_bso_id);
+  ephy_sync_service_queue_storage_request (self, endpoint,
+                                           SOUP_METHOD_DELETE,
+                                           NULL, -1, -1,
+                                           delete_open_tabs_record_cb, self);
+  g_free (endpoint);
+  g_free (device_bso_id);
 }
 
 static void
-ephy_sync_service_unregister_device (EphySyncService *self)
+ephy_sync_service_delete_client_record (EphySyncService *self)
 {
   char *endpoint;
-  char *id;
+  char *device_bso_id;
 
   g_assert (EPHY_IS_SYNC_SERVICE (self));
 
-  id = ephy_sync_utils_get_device_id ();
+  device_bso_id = ephy_sync_utils_get_device_bso_id ();
   /* Delete the client record associated to this device. */
-  endpoint = g_strdup_printf ("storage/clients/%s", id);
+  endpoint = g_strdup_printf ("storage/clients/%s", device_bso_id);
   ephy_sync_service_queue_storage_request (self, endpoint,
                                            SOUP_METHOD_DELETE,
                                            NULL, -1, -1,
-                                           unregister_device_cb, NULL);
+                                           delete_client_record_cb, self);
   g_free (endpoint);
-
-  /* Delete the open tabs record associated to this device. */
-  endpoint = g_strdup_printf ("storage/tabs/%s", id);
-  ephy_sync_service_queue_storage_request (self, endpoint,
-                                           SOUP_METHOD_DELETE,
-                                           NULL, -1, -1,
-                                           delete_open_tabs_record_cb, NULL);
-  g_free (endpoint);
-  g_free (id);
+  g_free (device_bso_id);
 }
 
 void
-ephy_sync_service_sign_out (EphySyncService *self,
-                            gboolean         unregister_device)
+ephy_sync_service_sign_out (EphySyncService *self)
 {
   g_return_if_fail (EPHY_IS_SYNC_SERVICE (self));
 
-  /* If we sign out without unregistering the device, then the current id of
-   * the device should not be cleared, but preserved for further use (Ephy will
-   * use the same id next time the user signs in). This way we prevent a zombie
-   * record in the clients collection on the Sync Storage Server.
-   */
-  if (unregister_device) {
-    ephy_sync_service_unregister_device (self);
-    ephy_sync_utils_set_device_id (NULL);
-  }
-
   ephy_sync_service_stop_periodical_sync (self);
-  ephy_sync_service_destroy_session (self, NULL);
-  ephy_sync_service_forget_secrets (self);
-  ephy_sync_service_clear_storage_queue (self);
-  ephy_sync_service_clear_storage_credentials (self);
+  ephy_sync_service_delete_client_record (self);
 
   /* Clear managers. */
   for (GSList *l = self->managers; l && l->data; l = l->next) {
@@ -2452,9 +2755,7 @@ ephy_sync_service_sign_out (EphySyncService *self,
   ephy_sync_utils_set_bookmarks_sync_is_initial (TRUE);
   ephy_sync_utils_set_passwords_sync_is_initial (TRUE);
   ephy_sync_utils_set_history_sync_is_initial (TRUE);
-
   ephy_sync_utils_set_sync_time (0);
-  ephy_sync_utils_set_sync_user (NULL);
 }
 
 void
